#coding:utf-8

"""
_____________                           .__ 
\__    ___/  | _____  ___          __ __|__|
  |    |  |  |/ /\  \/  /  ______ |  |  \  |
  |    |  |    <  >    <  /_____/ |  |  /  |
  |____|  |__|_ \/__/\_ \         |____/|__|
               \/      \/                   

The Tkx Python library, create modern frameless GUIs with Python and Tkinter.
    - author : Martin Gate <martingate98000@gmail.com>
    - github repo : https://www.github.com/MartinGate/tkx.git


Copyright (c) 2021, Martin Gate, Xvelta, Inc.
All rights reserved.
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the Xvelta, Inc. nor the
  names of its contributors may be used to endorse or promote products
  derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

try:
    import tkinter as tk # For Python 3.x
    from tkinter import TclError
except ImportError:
    import Tkinter as tk # For Python 2.x

# internal library imports
from .constants import *

# python built-in libraries imports
import threading
import platform

# Get the height of the taskbar for Windows users
if platform.system() == "Windows":
    try:
        from win32api import GetMonitorInfo, MonitorFromPoint

        monitor_info = GetMonitorInfo(MonitorFromPoint((0,0)))
        monitor_area = monitor_info.get("Monitor")
        work_area = monitor_info.get("Work")
        
        TOOLBAR_HEIGHT = monitor_area[3]-work_area[3]

    except ImportError or ModuleNotFoundError:
        print("!Tkx::WARNING -> You should install win32api by typing the following command : 'pip install win32api'.")
        TOOLBAR_HEIGHT = 0

# On other platforms, we consider that there is no taskbar
else:
    TOOLBAR_HEIGHT = 0


WIDTH_MARGIN, HEIGHT_MARGIN = 30, 10


class FramelessDragBehaviour():

    def __init__(self, master, widget, border_events=True):
        """
        Create a drag behaviour.

        You can apply it to a widget (both tkinter and ttk widgets) so that when you drag it,
        it will automatically move a specified window.

        For instance, the FramelessDragBehaviour class is used, 
        to simulate the border's drag events of a frameless window.

        Params:
            master : tkx.Tk / tkinter.Tk / tix.Tk / tkinter.TopLevel / tix.TopLevel [Draggable window]
            widget : Whatever widget used to drag the specified window
            border_events : True [default] / False [Should the FramelessDragBehaviour simulate the Windows' border resizing events]
        """
        self._master = master
        self._widget = widget
        self._border_events = border_events
        self.x = None
        self.y = None
        self.mouse_x, self.mouse_y = None, None

        self._widget.bind("<Button-1>", self.startMove)
        self._widget.bind("<ButtonRelease-1>", self.stopMove)
        self._widget.bind("<B1-Motion>", self.moving)

    def startMove(self, event=None):
        """ The widget starts moving. """
        self.x = event.x
        self.y = event.y

    def stopMove(self, event=None):
        """ The widget stops moving. """
        screen_width, screen_height = self._master.winfo_screenwidth(), self._master.winfo_screenheight()

        if self._border_events:
            # Reproduces window manager border-drag&drop fonctionnality
            if self.mouse_y <= HEIGHT_MARGIN and (self.mouse_x > WIDTH_MARGIN) and (self.mouse_x < (screen_width - WIDTH_MARGIN)):
                self._master._state = MAXIMIZED
                self._master.geometry("{}x{}+0+0".format(screen_width, screen_height - TOOLBAR_HEIGHT))

            elif self.mouse_y <= HEIGHT_MARGIN and self.mouse_x <= WIDTH_MARGIN:
                self._master._state = UPPER_LEFT
                self._master.geometry("{}x{}+0+0".format(screen_width//2, screen_height//2))

            elif self.mouse_y <= HEIGHT_MARGIN and self.mouse_x >= screen_width - WIDTH_MARGIN:
                self._master._state = UPPER_RIGHT
                self._master.geometry("{}x{}+{}+0".format(screen_width//2, screen_height//2, screen_width - screen_width//2))
            
            elif self.mouse_y >= screen_height - TOOLBAR_HEIGHT - HEIGHT_MARGIN and self.mouse_x <= WIDTH_MARGIN:
                self._master._state = LOWER_LEFT
                self._master.geometry("{}x{}+0+{}".format(screen_width//2, screen_height//2, screen_height - screen_height//2 - TOOLBAR_HEIGHT))
            
            elif self.mouse_y >= screen_height - TOOLBAR_HEIGHT - HEIGHT_MARGIN and self.mouse_x >= screen_width - WIDTH_MARGIN:
                self._master._state = UPPER_RIGHT
                self._master.geometry("{}x{}+{}+{}".format(screen_width//2, screen_height//2, screen_width - screen_width//2, screen_height - screen_height//2 - TOOLBAR_HEIGHT))

            elif self.mouse_x <= WIDTH_MARGIN:
                self._master._state = RIGHT
                self._master.geometry("{}x{}+0+0".format(screen_width//2, screen_height - TOOLBAR_HEIGHT))

            elif self.mouse_x >= screen_width - WIDTH_MARGIN:
                self._master._state = RIGHT
                self._master.geometry("{}x{}+{}+0".format(screen_width//2, screen_height - TOOLBAR_HEIGHT, screen_width - screen_width//2))

        # Refresh the window preview after 1 second
        change_window_preview_thread = threading.Thread(target=self._master._change_window_preview)
        self._master.after(1000, lambda: change_window_preview_thread.start())

        self.x = None
        self.y = None

    def moving(self, event=None):
        """ The widget is moving. """
        self.mouse_x, self.mouse_y = event.x_root, event.y_root

        x = (event.x_root - self.x - self._widget.winfo_x() + self._widget.winfo_x())
        y = (event.y_root - self.y - self._widget.winfo_y() + self._widget.winfo_y())

        self._master.geometry("+%s+%s" % (x, y))

    def disable_winbehaviours(self):
        self._border_events = False

    def enable_winbehaviours(self):
        self._border_events = True



#coding:utf-8

"""
_____________                           .__ 
\__    ___/  | _____  ___          __ __|__|
  |    |  |  |/ /\  \/  /  ______ |  |  \  |
  |    |  |    <  >    <  /_____/ |  |  /  |
  |____|  |__|_ \/__/\_ \         |____/|__|
               \/      \/                   

The Tkx Python library, create modern frameless GUIs with Python and Tkinter.
    - author : Martin Gate <martingate98000@gmail.com>
    - github repo : https://www.github.com/MartinGate/tkx.git


Copyright (c) 2021, Martin Gate, Xvelta, Inc.
All rights reserved.
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the Xvelta, Inc. nor the
  names of its contributors may be used to endorse or promote products
  derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

try:
    import tkinter as tk # For Python 3.x
    import tkinter.ttk as ttk
    from tkinter import font
except ImportError:
    import Tkinter as tk # For Python 2.x
    import ttk

# internal library imports
from .constants import *
from .behaviors import FramelessDragBehaviour

# external libraries imports
try:
    from PIL import ImageTk, ImageGrab
except ModuleNotFoundError:
    # print("!Tkx::ERROR -> You must install the Pillow library by typing the following command : 'pip install Pillow'.")
    raise ImportError("!Tkx::ERROR -> You must install the Pillow library by typing the following command : 'pip install Pillow'.")

# python built-in libraries imports
import threading


class CoreUI(tk.Tk):

    def __init__(self, *args, **kwargs):
        """
        Parent of the frameless window.
        The CoreUI class is used to display its child (the frameless window) on the taskbar,
        as it is normally ignored by the machine's window manager.
        """
        tk.Tk.__init__(self, **kwargs)
        self.title('tk')

        self.__preview = tk.Label(self)
        self.__preview.pack(fill=tk.BOTH, expand=1)

    def change_preview(self, img):
        """ Set a preview image to the frameless window. """
        self.__preview.config(image=img)
        self.__preview.image = img


class Tk(tk.Tk, tk.Toplevel):

    def __init__(self, display=NATIVE, defaultBorder=True, isResizable=True, **kwargs):
        """
        Create a tkx.Tk window.

        Params:
            display : tkx.NATIVE (normal window) [default] / tkx.FRAMELESS (frameless window, fully customizable)
            defaultBorder : True [default] / False (Wether to instanciate or not a default border for a frameless window)
            isResizable : True [default] / False (Wether to create or not a resizer object so that the frameless window can be resized)
            **kwargs : other tk.Tk parameters
        """
        self._display = display
        self._is_resizable = isResizable
        self._has_default_border = defaultBorder

        self.default_border_attributes = None

        self._state = NORMAL
        
        if self.is_frameless():
            self.__coreUI = CoreUI()
            self.__coreUI.attributes('-alpha', 0.0) #For icon
            self.__coreUI.bind("<Map>", self.deiconify_frameless)
            self.__coreUI.bind("<Unmap>", self.iconify_frameless)
            self.__coreUI.protocol("WM_DELETE_WINDOW", lambda : self.destroy())

            tk.Toplevel.__init__(self, self.__coreUI, **kwargs)

            self.overrideredirect(1) # Removes the border of the TopLevel
            self.attributes('-transparentcolor', TRANSPARENT)

            # Create a FramelessResizeBehaviour object to resize the window associated with a ttk.Sizer
            if self._is_resizable and self.is_frameless():
                self.resizer = ttk.Sizegrip(self)
                self.resizer.place(relx=1.0, rely=1.0, anchor="se")

                self.resizer.bind("<ButtonRelease-1>", self._change_window_preview)
                self.bind('<Map>', lambda a: self.tkraise_frameless())

            generate_preview = threading.Thread(target=self._change_window_preview)
            generate_preview.start()

        else:
            tk.Tk.__init__(self, **kwargs)

        # Make sure the window is on top of the CoreUI
        self.lift()

        if self._has_default_border:
            self.default_border_attributes = {
                    'close' : {
                        'bg': self.cget('bg'),
                        'fg': "black",
                        'hoverbg': "white",
                        'hoverfg': "black"
                        },
                    'minimize' : {
                        'bg': self.cget('bg'),
                        'fg': "black",
                        'hoverbg': "white",
                        'hoverfg': "black"
                    },
                    'maximize' : {
                        'bg': self.cget('bg'),
                        'fg': "black",
                        'hoverbg': "white",
                        'hoverfg': "black"
                    },
                    'border' : {
                        'bg': self.cget('bg')
                    }
                }
        
            self.createDefaultBorder()
    
    def current_state(self):
        return self._state

    def createDefaultBorder(self):
        """ Create a border for a frameless window if its defaultBorder parameter is set to True. """

        self.border = tk.Frame(self)
        self.border.config(bg=self.border.master.cget('bg'))

        self.border.framelessDragBehaviour = FramelessDragBehaviour(self, self.border)

        def on_minhover(event=None):
            self.border.minimize.config(bg=self.default_border_attributes['minimize']['hoverbg'])
            self.border.minimize.config(fg=self.default_border_attributes['minimize']['hoverfg'])

        def on_minunhover(event=None):
            self.border.minimize.config(bg=self.default_border_attributes['minimize']['bg'])
            self.border.minimize.config(fg=self.default_border_attributes['minimize']['fg'])

        def on_maxhover(event=None):
            self.border.maximize.config(bg=self.default_border_attributes['maximize']['hoverbg'])
            self.border.maximize.config(fg=self.default_border_attributes['maximize']['hoverfg'])

        def on_maxunhover(event=None):
            self.border.maximize.config(bg=self.default_border_attributes['maximize']['bg'])
            self.border.maximize.config(fg=self.default_border_attributes['maximize']['fg'])

        def on_closehover(event=None):
            self.border.close.config(bg=self.default_border_attributes['close']['hoverbg'])
            self.border.close.config(fg=self.default_border_attributes['close']['hoverfg'])

        def on_closeunhover(event=None):
            self.border.close.config(bg=self.default_border_attributes['close']['bg'])
            self.border.close.config(fg=self.default_border_attributes['close']['fg'])

        self.border.close = tk.Button(self.border, text = "X", cursor="hand2", bg=self.default_border_attributes['close']['bg'], fg=self.default_border_attributes['close']['fg'], border=0, command = lambda: self.destroy(), font=("Consolas", 15))
        self.border.close.bind("<Enter>", on_closehover)
        self.border.close.bind("<Leave>", on_closeunhover)
        self.border.close.pack(side=tk.RIGHT)

        char = u"\u25A1"
        self.border.maximize = tk.Button(self.border, text = char, cursor="hand2", bg=self.default_border_attributes['maximize']['bg'], fg=self.default_border_attributes['maximize']['fg'], border=0, command = lambda: self.maximize(), font=("Consolas", 15))
        self.border.maximize.bind("<Enter>", on_maxhover)
        self.border.maximize.bind("<Leave>", on_maxunhover)
        self.border.maximize.pack(side=tk.RIGHT, padx=(0, 2))

        self.border.minimize = tk.Button(self.border, text = "_", cursor="hand2", bg=self.default_border_attributes['minimize']['bg'], fg=self.default_border_attributes['minimize']['fg'], border=0, command = lambda: self.withdraw(), font=("Consolas", 15))
        self.border.minimize.bind("<Enter>", on_minhover)
        self.border.minimize.bind("<Leave>", on_minunhover)
        self.border.minimize.pack(side=tk.RIGHT, padx=(0, 2))

        self.border.pack(fill=tk.X, anchor='nw')

    def maximize(self, event=None):
        """ Maximize the window. """
        self.geometry("{}x{}+0+0".format(self.winfo_screenwidth(), self.winfo_screenheight() - 40))

    def tkraise_frameless(self, event=None):
        """ Raise the window resizer always on top of other widgets """
        self.resizer.tkraise()

    def resize_core(self, event=None):
        """
        Resize the parent window of the frameless window to fit the size of its child.
        (used to display the window in the taskbar
        despite the overrideredirect flag set to the frameless window)
        """
        self.__coreUI.geometry("{}x{}".format(self.winfo_width(), self.winfo_height()))
        
    def _change_window_preview(self, event=None):
        """
        Generate a preview (displayed in the taskbar) of a frameless window.
        """
        self.resize_core()
  
        bbox = (self.winfo_x(), self.winfo_y(), self.winfo_x() + self.winfo_width(), self.winfo_y() + self.winfo_height())

        img = ImageGrab.grab(bbox)
        self.__coreUI.change_preview(ImageTk.PhotoImage(img))

    def deiconify_frameless(self, event=None):
        """ Internal method to make the frameless window visible when its parent is mapped. """
        self.deiconify()
    
    def iconify_frameless(self, event=None):
        """ Internal method to make the frameless window invisible when its parent is unmapped. """
        self.withdraw()

    def iconbitmap(self, icon):
        """ Change the icon of the window. """
        if self.is_frameless():
            self.__coreUI.iconbitmap(icon)
        else:
            tk.Tk.iconbitmap(self, icon)
    
    def title(self, title="tkx"):
        """ Change the title of the window. """
        if self.is_frameless():
            self.__coreUI.title(title)
        else:
            tk.Tk.title(self, title)

    def is_frameless(self):
        """ Check if the window is a frameless window or not. """
        if self._display == FRAMELESS:
            return True
        else:
            return False

    def destroy(self):
        """ Destroy the window. """ 
        if self.is_frameless():
            tk.Toplevel.destroy(self)
            self.__coreUI.destroy()
        else:
            tk.Tk.destroy(self)

    def attributes(self, attr, *value):
        """ Setter:
                Set an attribute to the window.
            Getter:
                Get an attribute of the window.
        """
        if self.is_frameless():
            if attr == '-topmost':
                if value != ():
                    tk.Toplevel.attributes(self, attr, *value)

                    if tk.Toplevel.attributes(self, '-topmost'):
                        self.__coreUI.unbind("<Unmap>")
                    else:
                        self.__coreUI.bind("<Unmap>", self.iconify_frameless)

                else:
                    return tk.Toplevel.attributes(self, attr)

            else:
                if value != ():
                    tk.Toplevel.attributes(self, attr, *value)
                else:
                    return tk.Toplevel.attributes(self, attr)

        else:
            if value != ():
                tk.Tk.attributes(self, attr, *value)
            else:
                return tk.Tk.attributes(self, attr)

    def center(self):
        """ Center the window in the center of the screen. """
        self.update_idletasks()
        # Get the requested values of the height and width.
        windowWidth = self.winfo_width()
        windowHeight = self.winfo_height()
        
        # Get both half the screen width/height and window width/height
        positionRight = int(self.winfo_screenwidth()/2 - windowWidth/2)
        positionDown = int(self.winfo_screenheight()/2 - windowHeight/2)
        
        # Positions the window in the center of the page.
        self.geometry("+{}+{}".format(positionRight, positionDown))

    def disable_winbehaviours(self, event=None):
        """ Disable the Windows borders drag&drop-resize events. """
        if self.is_frameless() and self._has_default_border:
            self.border.framelessDragBehaviour._border_events = False
    
    def enable_winbehaviours(self, event=None):
        """ Enable the Windows borders drag&drop-resize events. """
        if self.is_frameless() and self._has_default_border:
            self.border.framelessDragBehaviour._border_events = True

    def resizer_bg(self, bg="#2A2A2A"):
        """ Change the background color of the frameless window's sizer. """
        style = ttk.Style()
        style.configure("TSizegrip", background=bg)

    def config_border(self, config):
        """
        Configure the attributes of the default border's widgets.

        Example : To make the background color of the close button red and its text color black, the argument must be :
            {
                'close': {'bg':"red", 'fg':"black"}
            }

        You can also pass extra args for the widgets' hover colors, for instance:
            'minimize' : {'hoverbg': "lightblue", 'hoverfg': "white"}


        Widgets you can configure : "close", "minimize", "border" (doesn't support hover colors), "maximize"
        """
        if self._has_default_border:
            for wdgt in config:
                for attr in config[wdgt]:
                    if attr == "hoverbg" or attr == "hoverfg":
                        pass
                    else:
                        if wdgt == 'close':
                            widget = self.border.close
                        if wdgt == 'minimize':
                            widget = self.border.minimize
                        if wdgt == 'border':
                            widget = self.border
                        if wdgt == 'maximize':
                            widget = self.border.maximize
                        
                        temp_attr = {attr: config[wdgt][attr]}
                        widget.config(temp_attr)

                    self.default_border_attributes[wdgt][attr] = config[wdgt][attr]
        else:
            pass


if __name__ == '__main__':
    root = Tk(display=FRAMELESS)
    root.geometry("500x300")
    root.disable_winbehaviours()
    root.center()
    root.mainloop()



#coding:utf-8

"""
_____________                           .__ 
\__    ___/  | _____  ___          __ __|__|
  |    |  |  |/ /\  \/  /  ______ |  |  \  |
  |    |  |    <  >    <  /_____/ |  |  /  |
  |____|  |__|_ \/__/\_ \         |____/|__|
               \/      \/                   

The Tkx Python library, create modern frameless GUIs with Python and Tkinter.
    - author : Martin Gate <martingate98000@gmail.com>
    - github repo : https://www.github.com/MartinGate/tkx.git


Copyright (c) 2021, Martin Gate, Xvelta, Inc.
All rights reserved.
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the Xvelta, Inc. nor the
  names of its contributors may be used to endorse or promote products
  derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from .constants import *
from .behaviors import FramelessDragBehaviour
from .Tk import Tk
from .JSONLoader import JSONLoader



#coding:utf-8

"""
_____________                           .__ 
\__    ___/  | _____  ___          __ __|__|
  |    |  |  |/ /\  \/  /  ______ |  |  \  |
  |    |  |    <  >    <  /_____/ |  |  /  |
  |____|  |__|_ \/__/\_ \         |____/|__|
               \/      \/                   

The Tkx Python library, create modern frameless GUIs with Python and Tkinter.
    - author : Martin Gate <martingate98000@gmail.com>
    - github repo : https://www.github.com/MartinGate/tkx.git


Copyright (c) 2021, Martin Gate, Xvelta, Inc.
All rights reserved.
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the Xvelta, Inc. nor the
  names of its contributors may be used to endorse or promote products
  derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

try:
    import tkinter as tk # For Python 3.x
    import tkinter.ttk as ttk
except ImportError:
    import Tkinter as tk # For Python 2.x
    import ttk

# internal library imports
from tkinter import TclError, constants
from .constants import *
from .behaviors import FramelessDragBehaviour
from .Tk import Tk

# python built-in modules imports
import json


class JSONLoader():

    def __init__(self, master, json_file=None):
        self.master = master
        self.__ui_elements = {}
        self.__custom_widgets = {}
        self.__classes = {}
        self.__names = {}

        self.generate(json_file)

    def generate(self, json_file=None):
        if json_file == None:
            return

        # load Json data
        with open(json_file , 'r') as file:
            data = file.read()
            data_list = data.split('"')

            counter = 0

            for i in range(len(data_list)):
                if data_list[i][:2] == "tk" or data_list[i][:3] == "ttk" or data_list[i][0] == "@":
                    data_list[i] = data_list[i] + '#' + str(counter)

                    counter += 1

            data = '"'.join(data_list)
            self.__ui_elements[json_file] = json.loads(data)

        if self.__ui_elements[json_file]['doctype'] == 'ui-definition':
            # Create a queue of to-create widgets
            widget_queue = self.create_widget_queue(self.__ui_elements[json_file])

            # Create each widget
            for zone in widget_queue:
                for widget in zone:
                    if widget['zone'] == "border":
                        parent = "self.master.border"
                    elif widget['zone'] == 'main':
                        parent = "self.master"
                    else:
                        # A Custom Widget Handler object is its parent
                        try:
                            parent = "self.master." + widget['zone']
                        except TypeError:
                            parent = "self.master"
                    
                    widget_type = widget['widget_type'].split('#')[0]
                    config = widget["config"]

                    if widget["parent"] == '':
                        draw_parent = "self.master"
                    else:
                        draw_parent = "self.master." + widget["parent"]

                    self.create_widget(parent, widget_type, config.copy(), draw_parent=draw_parent)

        elif self.__ui_elements[json_file]['doctype'] == 'theme':
            self.__applyTheme(self.__ui_elements[json_file])
        
    def create_widget_queue(self, json_data):
        widget_queue = []

        def check_subwidgets(widget_attrs, zone=None):
            sub_widgets = []
            for i in widget_attrs:
                if i[:2] == "tk" or i[:3] == "ttk" or i[0] == "@":
                    try:
                        parent = widget_attrs['name']
                    except KeyError:
                        if zone == 'border':
                            parent = zone
                        else:
                            parent = ''

                    if i[0] != "@":
                        sub_widgets.append({'widget_type':i, 'config':widget_attrs[i], 'parent':parent, 'zone': zone})

                        temp_sub_widgets = check_subwidgets(widget_attrs[i])
                        try:
                            for i in range(len(temp_sub_widgets)):
                                sub_widgets.append(temp_sub_widgets[i])
                        except TypeError:
                            # Object does not have subwidgets
                            pass

                    else:
                        class CustomWidgetHandler(object):
                            def __init__(self):
                                """ Object used to handle a custom widget. """
                                pass

                        # Create a custom widget handler to get access to the subwidgets
                        try:
                            eval(f"self.master.{widget_attrs[i]['name']}")
                        except AttributeError:
                            # Create the custom widget handler
                            try:
                                exec(f"self.master.{widget_attrs[i]['name']} = CustomWidgetHandler()")
                            except Exception:
                                print("!Tkx::ERROR: Unable to create custom widget handler.")

                        # There is a custom widget
                        widget_name = "widget:"+ i.split("@")[1].split("#")[0]

                        for n in range(len(self.__custom_widgets[widget_name])):
                            dict_custom_widget = self.__custom_widgets[widget_name][n].copy()

                            if dict_custom_widget['parent'] != '':
                                dict_custom_widget['parent'] = widget_attrs[i]['name'] + '.' + dict_custom_widget['parent']
                            else:
                                dict_custom_widget['parent'] = parent
                            
                            dict_custom_widget['zone'] = widget_attrs[i]['name']                            
                            sub_widgets.append(dict_custom_widget)
            
            if sub_widgets: 
                return sub_widgets

        # The user is not allowed to put the same zone twice, so we use a for loop
        for zone in json_data:
            # Set the exploring zone
            if zone == 'main' or zone == 'border':
                widget_queue.append(check_subwidgets(json_data[zone], zone=zone))
            elif zone[:6] == 'widget':
                self.__custom_widgets[zone] = check_subwidgets(json_data[zone], zone=zone)

        return widget_queue

    def __applyTheme(self, data):
        def all_children (wid) :
            _list = wid.winfo_children()

            for item in _list :
                if item.winfo_children() :
                    _list.extend(item.winfo_children())

            return _list

        def apply_attr(widget, data):
            try:
                widget.config(data)
            except TclError as e:
                print(f"!Tkx::ERROR: Cannot apply theme to {type(widget).__name__} -->", e)

        children = all_children(self.master)
        
        for i in children:
            for attr in data:
                # Get name attribute
                try:
                    name = attr.split('#')[1]
                except Exception:
                    name = "NONAME"
                
                # Get class attribute
                try:
                    _class = attr.split('.')[1]
                except Exception:
                    _class = 'NOCLASS'

                _attr = attr.split('#')[0]
                    
                if "tk." + type(i).__name__ == _attr:
                    apply_attr(i, data[attr])

                if attr == "doctype":
                    pass
                if attr == "window":
                    apply_attr(self.master, data[attr])
                if attr == "border":
                    try:
                        self.master.config_border(data[attr])
                    except Exception:
                        print("!Tkx:WARNING --> You can't customize the border of a normal tkinter window, you should use a Tkxui window instead.")

                if name in self.__names:
                    apply_attr(self.__names[name], data[attr])
                
                if _class in self.__classes:
                    print(_class)
                
                    for widget in self.__classes[_class]:
                        apply_attr(widget, data[attr])
               
    def create_widget(self, parent_widget, widget_type, widget_dict, draw_parent="self.master"):
        # Remove all sub_widgets
        sub_widgets = []
        for attr in widget_dict:
            if attr[:2] == "tk" or attr[:3] == "ttk" or attr[0] == "@":
                sub_widgets.append(attr)

        for to_remove in sub_widgets:
            del widget_dict[to_remove]
        
        try:
            # Get the given name of the widget 
            widget_name =  widget_dict['name']
            # Delete this non-recognize attribute
            del widget_dict['name']

        except KeyError:
            # if no one is specified, the user won't be able to access it from its parent
            widget_name = "unnamed"

        # Get the pack options of the widget
        try:
            pack_options = widget_dict['pack_options']
            # Retrieve the non-recognize attribute
            del widget_dict['pack_options']

        except KeyError:
            pack_options = ""

        # Get the class of the widget
        try:
            widget_class = widget_dict['class']
            # Retrieve the non-recognize attribute
            del widget_dict['class']
        except KeyError:
            widget_class = None
        
        # Create the widget
        try:
            exec(f"{parent_widget}.{widget_name} = {widget_type}({draw_parent})")
            
            if widget_class != None:
                # Assigning its class
                try:
                    # Get the list of widget that have the same class
                    other_widgets = self.__classes[widget_class]
                except:
                    other_widgets = []

                other_widgets.append(eval(f"{parent_widget}.{widget_name}"))
                self.__classes[widget_class] = other_widgets 

            if widget_name != "unnamed":
                self.__names[widget_name] = eval(f"{parent_widget}.{widget_name}")

            temp_widget = eval(f"{parent_widget}.{widget_name}")
            temp_widget.config(**widget_dict)
        except Exception as e:
            print(f"!Tkx::WARNING: Unable to create {widget_type} ---> " + str(e))

        # Pack the widget
        try:
            for attr in pack_options:
                if pack_options[attr][0] == '(':
                    pack_options[attr] = eval(pack_options[attr])

            exec(f"{parent_widget}.{widget_name}.pack({pack_options})")

        except Exception as e:
            print(f"!Tkx::WARNING: Unable to pack {widget_type} ---> " + str(e))

                
if __name__ in '__main__':
    root = Tk(display=FRAMELESS)
    root.geometry("800x500")
    root.center()
    json_loader = JSONLoader(root)
    json_loader.generate("ui.json")
    json_loader.generate("dark_theme.json")

    # root.disable_winbehaviours()
    root.config_border({
        'close':{
            'hoverbg': 'red',
            'hoverfg': "white"
        }
    })

    # root.label_frame_container.config(bg="red")

    root.mainloop()
        



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tkxui",
    version="1.0.0",
    author="Martin Gate",
    author_email="martingate98000@gmail.com",
    description="Create modern frameless GUIs with Python, Tkinter and JSON",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/GateMartin/TkXUI",
    project_urls={
        "Bug Tracker": "https://github.com/GateMartin/TkXUI/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.2",
)